import os
import base64
import socket
import ldap
import ldap.sasl

try:
    from gssapi import creds as gsscreds
    from gssapi import sec_contexts as gssctx
    from gssapi import names as gssnames
    with_gssapi = True
except:
    with_gssapi = False

from flask import Blueprint, request, Response
from flask.helpers import Headers

from functools import wraps

from lib389 import DirSrv
from lib389._constants import *

from .Helpers import jsonResponse

dir_srv_conn = None

auth_api = Blueprint('auth_api', __name__)


def getConnection(binddn='', password='', saslmethod=None):
    """
    Get the DS connection object
    """
    dir_srv_conn = DirSrv(verbose=False)
    headers = request.headers
    inst = {}

    print(headers)
    #
    # Parse the headers looking for host name
    #
    # We need to prefix these with LDAP, because host is reserved for HTTP
    # things.
    inst[SER_HOST] = headers.get('ldaphost', socket.gethostname())

    try:
        inst[SER_PORT] = int(headers.get('ldapport', '389'))
    except ValueError:
        inst[SER_PORT] = 389

    inst[SER_SERVERID_PROP] = headers.get('ldapinst', None)

    # We need apache to determine this, not us. Set to simple
    # saslmethod = str(headers.get('ldapsaslmethod', None))

    # Get certificate database directory name
    # don't wrap this in str, else bad things happen :)
    certdir = headers.get('ldapcertdir', None)

    inst[SER_ROOT_DN] = binddn
    inst[SER_ROOT_PW] = password
    dir_srv_conn.allocate(inst)
    dir_srv_conn.open(saslmethod, certdir)

    return dir_srv_conn


def _negotiate_start(req):
    """
    Start Negotiation
    """
    # We don't need to specify the realm name, gssapi works it out for us.
    svc_princ = gssnames.Name('HTTP/%s' % (socket.gethostname()))
    server_creds = gsscreds.Credentials(usage='accept', name=svc_princ)
    context = gssctx.SecurityContext(creds=server_creds)

    # Yay! Undocumented gssapi magic. No indication that you need to b64 decode
    context.step(base64.b64decode(req))
    deleg_creds = context.delegated_creds
    # WARNING: This only works because lib389 is in the same process / memory
    # space. WE CANNOT USE THESE CREDS IN A CALL TO OTHER APPLICATIONS.
    CCACHE = 'MEMORY:ccache_rest389_%s' % deleg_creds.name
    store = {'ccache': CCACHE}
    deleg_creds.store(store, overwrite=True)
    # We need to set this in the env so that py.ldap can find it
    os.environ['KRB5CCNAME'] = CCACHE
    # Return the context, so we can free it later.
    return context


def _negotiate_end(context):
    """
    End Negotiation
    """
    # We still need to destroy the creds though if WE set them.
    # When we throw away the gss context, this memory cred is destroyed
    # but there is no way in python-gssapi to free gss_cred_id_t
    deleg_creds = context.delegated_creds
    del(deleg_creds)
    # del(context)


def _connection(f, *args, **kwargs):
    """
    Open a connection
    """
    retval = None
    auth = request.authorization
    negotiate = False
    dir_srv_conn = False
    binddn = None
    password = None
    headers = Headers()  # Allows a multivalue header response.

    authorization = request.headers.get("Authorization", None)

    try:
        # First, try to build a gssapi context
        if authorization is not None and with_gssapi:
            values = authorization.split()
            if values[0] == 'Negotiate':
                # Now we really start
                # If this is valid, it sets KRB5CCNAME
                negotiate = _negotiate_start(values[1])

        # This is provided by mod_auth_gssapi. If it's there use it, but it's
        # not optimal for our use case.
        if request.headers.get("Krb5Ccname", '(null)') != '(null)':
            os.environ['KRB5CCNAME'] = request.headers.get("Krb5Ccname", None)

        if os.environ.get('KRB5CCNAME', '') != '':
            try:
                dir_srv_conn = getConnection(saslmethod='GSSAPI')
            except ldap.INVALID_CREDENTIALS:
                # Dir_srv_conn stays false, so we give a 401 later
                pass
        else:
            if auth and (
                    auth.username is not None and auth.password is not None):
                binddn = auth.username
                password = auth.password
            # Check if the username is a principal.
            if binddn is not None and password is not None:
                try:
                    dir_srv_conn = getConnection(binddn, password)
                except ldap.INVALID_CREDENTIALS:
                    # DS will still be false, so we can just pass
                    pass
        if dir_srv_conn is False:
            headers.set('WWW-Authenticate', 'Basic realm="Admin"')
            if with_gssapi:
                headers.add('WWW-Authenticate', 'Negotiate')
            return Response("Unauthorized", 401, headers)
        retval = f(dir_srv_conn, *args, **kwargs)
    finally:
        # Close our DS connection/auth after usage, else we leak creds via
        # threads.
        if dir_srv_conn is not False:
            dir_srv_conn.close()
        if negotiate is not False:
            # This should invalidate the ccache contents.
            _negotiate_end(negotiate)
        if os.environ.get('KRB5CCNAME', None) is not None:
            # overwrite this to at least stop internal cred leaks.
            os.environ['KRB5CCNAME'] = ''
    return retval


# We can't do a callable / non-callable decorator due to flask magic.
def authenticateConnection(f):
    """
    Authenticate the connection
    """
    @wraps(f)
    def decorator(*args, **kwargs):
        """
        Connection decorator
        """
        return _connection(f, *args, **kwargs)
    return decorator


@auth_api.route('/v1/whoami', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def whoami(ds):
    """
    Returns the DN of the current user bound to the DS from your authentication
    details, or will return "" for Anonymous bind.
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {'allow': 'GET OPTIONS HEAD'}
        return jsonify(msg)

    if request.method == 'HEAD':
        try:
            whoami = ds.whoami_s()
            return jsonResponse("", 200, request.url)
        except ldap.LDAPError as e:
            return jsonResponse("", 403, request.url)

    if request.method == 'GET':
        # Return the identity of who is bound to ldap right now.
        # This returns a string, so maybe there is a better way to return this?
        try:
            whoami = ds.whoami_s()
            return jsonResponse(whoami, 200, request.url)
        except ldap.LDAPError as e:
            msg = 'Error: whoami failed (%s)' % str(e)
            return jsonResponse(msg, 403, request.url)
