import ldap
from ldap import filter as ldap_filter
from ldap.controls.readentry import PostReadControl
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
#from lib389.utils import *
from flask import request, jsonify
from flask import Blueprint

from .Authentication import authenticateConnection
from .Helpers import jsonResponse

suffix_api = Blueprint('suffix_api', __name__)

"""
Suffix Representation::

    {
        'dn': Config DN,
        'suffix': 'ou=people,dc=example,dc=com',
        'parent': 'dc=example,dc=com',
        'state':  'backend',
        'backend':  [ BACKEND, BACKEND, ... ],
        'referral': [ LDAP URL, LDAP URL, ... ]
    }
"""

SUFFIX = 'cn'
PARENT = 'nsslapd-parent-suffix'
STATE = 'nsslapd-state'
BACKEND = 'nsslapd-backend'
REFERRAL = 'nsslapd-referral'

pr = PostReadControl(criticality=True, attrList=['*'])


def jsonReturnSuffixes(suffixes, status_code, link):
    # Return our suffix reps
    results = []
    for suffix in suffixes:
        suffix_rep = {}
        suffix_rep['dn'] = suffix.dn
        suffix_rep['suffix'] = suffix.getValue(SUFFIX)
        if not suffix.getValue(PARENT):
            suffix_rep['parent'] = ""
        else:
            suffix_rep['parent'] = suffix.getValue(PARENT)
        suffix_rep['state'] = suffix.getValue(STATE)
        suffix_rep['backend'] = suffix.getValues(BACKEND)
        suffix_rep['referral'] = suffix.getValues(REFERRAL)
        results.append(dict(suffix_rep))

    return jsonResponse(results, status_code, link)


@suffix_api.route('/v1/suffix', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def suffixList(ds):
    """
    Returns a list of suffixes, or backends, avaliable on the server.
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    if request.method == 'HEAD':
        #
        # Return HEAD response
        #
        return jsonResponse("", 200)

    if request.method == 'GET':
        #
        # Return each suffix entry
        #
        try:
            filter_str = 'objectClass=nsMappingTree'
            entries = ds.search_s("cn=mapping tree,cn=config",
                                  ldap.SCOPE_SUBTREE, filter_str)
            return jsonReturnSuffixes(entries, 200, request.url)
        except ldap.LDAPError as e:
            msg = ('Failed to find any suffixes: ' + str(e))
            return jsonResponse(msg, 403)


@suffix_api.route('/v1/suffix/<suffix>', methods=['GET', 'PATCH', 'PUT',
                                                  'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def suffixResource(ds, suffix):
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET PATCH PUT DELETE OPTIONS HEAD'
        return jsonify(msg)

    # An example of ldap filter escaping.
    suffix = ldap_filter.escape_filter_chars(suffix)
    # We can use this entry for GET and PUT
    try:
        filter_str = ('(&(cn=%s)(objectclass=nsMappingTree))' % (suffix))
        entry = ds.search_s("cn=mapping tree,cn=config",
                            ldap.SCOPE_SUBTREE, filter_str)
    except ldap.LDAPError as e:
        pass

    if request.method == 'HEAD':
        #
        # Return HEAD response
        #
        if entry:
            return jsonResponse("", 200, request.url)
        else:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        #
        # Get suffix entries
        #
        try:

            #
            # Generate the response and resource link
            #
            if entry:
                return jsonReturnSuffixes(entry, 200, request.url)
            else:
                msg = ('Failed to find any suffixes')
                return jsonResponse(msg, 404, request.url)
        except ldap.LDAPError as e:
            msg = ('Failed to search for suffixes: ' + str(e))
            return jsonResponse(msg, 403, request.url)

    elif request.method == 'PUT':
        #
        # parse json
        #
        json_req = request.get_json(force=True)

        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                403, request.url)

        #
        # Add suffix
        #
        if 'backend' in json_req:
            backend = json_req['backend']
            #
            # Create the backend, okay if it already exists
            #
            try:
                ds.backend.create(suffix, {BACKEND_NAME: str(backend)})
            except ldap.ALREADY_EXISTS:
                # This is okay
                pass
            except ldap.LDAPError as e:
                return jsonResponse('%s' % (str(e)), 403, request.url)

            #
            # Create the mapping tree/suffix
            #
            try:
                if 'parentdn' in json_req:
                    parentdn = json_req['parentdn']
                else:
                    parentdn = None
                suffix_entry = ds.mappingtree.create(suffix,
                                                     bename=str(backend),
                                                     parent=parentdn)
                return jsonReturnSuffixes([suffix_entry], 201, request.url)
            except ldap.ALREADY_EXISTS:
                # idempotent - return an entry rep
                return jsonReturnSuffixes([entry], 201, request.url)
            except ldap.LDAPError as e:
                return jsonResponse('Failed to create suffix: %s ' %
                                    str(e), 403, request.url)
            except ValueError as e:
                return jsonResponse('Failed to create suffix: %s ' %
                                    str(e), 403, request.url)
        else:
            return jsonResponse('Missing "backend" element in JSON request',
                                403, request.url)

    elif request.method == 'PATCH':
        #
        # Modify/patch the suffix representation
        #
        mod_list = []
        json_rep = request.get_json(force=True)
        for mod in json_rep:
            # Get the op (add, replace, delete) and parse the path
            op = jsonGetPatchOp(mod)
            if not op:
                return jsonResponse(('invalid PATCH op: (%s)' % mod['op']),
                                    400, request.url)

            if mod['path'] == 'suffix':
                attr = SUFFIX
            elif mod['path'] == 'parent':
                attr = PARENT
            elif mod['path'] == 'state':
                attr = STATE
            elif mod['path'] == 'backend':
                attr = BACKEND
            elif mod['path'] == 'referral':
                attr = REFERRAL
            else:
                # Some other attribute?
                attr = mod['path']

            if isinstance(mod['value'], list):
                value = list(map(str, mod['value']))
            else:
                value = str(mod['value'])
            mod_list.append((op, attr, value))

        #
        # Perform the modifications
        #
        try:
            suffix_entry = ds.mappingtree.list(suffix)[0]
            msg_id = ds.modify_ext(suffix_entry.dn, mod_list, serverctrls=[pr])
        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

        #
        # Generate the response and resource link
        #
        try:
            _, _, _, resp_ctrls = ds.result3(msg_id)
        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

        entry = resp_ctrls[0].entry
        if REFERRAL in entry:
            ref = entry[REFERRAL]
        else:
            ref = ''
        if PARENT in entry:
            parent = entry[PARENT]
        else:
            parent = ''

        suffix_rep = {'dn': resp_ctrls[0].dn,
                      'suffix': entry[SUFFIX],
                      'backend': entry[BACKEND],
                      'referral': ref,
                      'state': entry[STATE],
                      'parent': parent}
        link = '%sv1/suffix/%s' % (request.url_root, suffix)

        return jsonResponse(suffix_rep, 200, link)

    elif request.method == 'DELETE':
        #
        # Get the backend name
        #
        backend = suffixToBackend(ds, suffix)
        if not backend:
            return jsonResponse('Failed to locate backend name', 403,
                                request.url)

        #
        # Delete the mapping tree(suffix)
        #
        try:
            ds.mappingtree.delete(suffix=suffix)
        except ldap.NO_SUCH_OBJECT:
            # idempotent - continue
            pass
        except ldap.LDAPError as e:
            msg = 'backend (%s) suffix (%s)' % (backend, suffix)
            return jsonResponse(msg, 403, request.url)

        #
        # Delete the backend
        #
        try:
            ds.backend.delete(suffix, bename=backend)
        except ldap.NO_SUCH_OBJECT:
            # idempotent
            pass
        except Excaption as e:
            return jsonResponse('%s' % str(e), 403, request.url)

        return jsonResponse('Deleted suffix and backend (%s).' %
                            backend, 200, request.url)
